;;; idm-mode-el -- Major mode for editing IDM files

;; Author: Jim Hague <jim.hague@acm.org>
;; Created: 13th January 2006
;; Keywords: IDM major-mode

;; Copyright (C) 2006 Jim Hague <jim.hague@acm.org>

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2 of
;; the License, or (at your option) any later version.

;; This program is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY; without even the implied
;; warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
;; PURPOSE.  See the GNU General Public License for more details.

;; You should have received a copy of the GNU General Public
;; License along with this program; if not, write to the Free
;; Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
;; MA 02111-1307 USA

;;; Commentary:
;; 
;; This mode is based on an example used in a tutorial about Emacs
;; mode creation. The tutorial can be found here:
;; http://two-wugs.net/emacs/mode-tutorial.html
;;
;; The indentation is based on jam.el.
;;
;; To use this mode, add
;;
;; (require 'idm-mode)
;;
;; to your .emacs or .xemacs/init.el. For automatic indentation as you
;; type, also add
;;
;; (defun my-idm-mode-hook ()
;;   (define-key idm-mode-map "\C-m" 'idm-electric-newline)
;; )
;; (add-hook 'idm-mode-hook 'my-idm-mode-hook)
;;
;; Every time you type 'Enter' this will re-indent the current line
;; add a new line after it and move to the current indent in that line.
 

;;; Code:
(defvar idm-mode-hook nil)

(defvar idm-mode-map
  (let ((idm-mode-map (make-keymap)))
    ; (define-key idm-mode-map "\C-m" 'newline-and-indent)
    idm-mode-map)
  "Keymap for IDM major mode")

(add-to-list 'auto-mode-alist '("\\.idm\\'" . idm-mode))
(add-to-list 'auto-mode-alist '("\\.mod\\'" . idm-mode))
(add-to-list 'auto-mode-alist '("\\.if\\'" . idm-mode))

(defconst idm-font-lock-keywords-1
  (list
   ;; keywords
   '("\\<\\(case\\|endcase\\|if\\|elseif\\|endif\\|for\\|endfor\\|while\\|endwhile\\|do\\|then\\|else\\|constant\\|variable\\|this\\|print\\|and\\|or\\|not\\|true\\|false\\|nothing\\|return\\)\\>" .
     font-lock-keyword-face))
  "Minimal keyword highlighting for IDM mode.")

(defconst idm-font-lock-keywords-2
  (append idm-font-lock-keywords-1
     (list
       ;; functions, variables, models, rules etc
       (list "\\<\\(record\\|default\\|function\\|model\\|child\\|rule\\|timer\\|module\\|import\\|dialog\\|export\\|format\\|observer\\)\\>[ \t]+\\([a-zA-Z_0-9-]+\\>\\)[ \t]*\\([a-zA-Z_0-9-]*\\).*$"
         '(1 font-lock-keyword-face)
         '(2 font-lock-type-face)
         '(3 font-lock-function-name-face))))
  "Additional keywords and function name highlighing for IDM mode.")

(defconst idm-font-lock-keywords-3
  (append idm-font-lock-keywords-2
    (list
      ;; types
      (list "\\(^\\|[ \t]+\\)\\<\\(object\\|pointer\\|integer\\|anyvalue\\|attribute\\|boolean\\|datatype\\|enum\\|event\\|index\\|integer\\|string\\|color\\|tile\\|cursor\\|font\\|accelerator\\|text\\)\\>[ \t]+\\([a-zA-Z_0-9-]+\\>\\)"
        '(1 font-lock-type-face)
        '(2 font-lock-variable-name-face))))
  "Types highlighting for IDM mode.")

(defconst idm-font-lock-keywords-4
  (append idm-font-lock-keywords-3
    (list
      ;; on ... rules
      (list "\\<\\(on\\)\\>[ \t]+\\([a-zA-Z_0-9._-]+\\>\\)[ \t]*\\([a-zA-Z_0-9._-]*\\)[ \t]+\\([a-zA-Z_0-9_-]+\\>\\).*$"
        '(1 font-lock-keyword-face)
        '(2 font-lock-function-name-face)
        '(3 font-lock-function-name-face)
        '(4 font-lock-type-face))

      ;; Case possibilities - strings screw up regexps so second in is for
      ;; in "...":
      '("\\<\\(in\\>[ \t]+\\<.*\\>:\\|otherwise:\\)" .
        font-lock-reference-face)

      '("\\<in\\>" . font-lock-reference-face)))
  "Full max welly IDM highlighting.")

(defvar idm-font-lock-keywords idm-font-lock-keywords-4
  "Default highlighting expressions for IDM mode.")

;; Line indenting.
;;
;; The line indenting here is very simplistic. It will only work when
;; code blocks are started and terminated by the first item on a line.
;; It doesn't attempt any sort of parsing of the program, and so is
;; vulnerable to continuation lines, string text etc.

(defvar idm-indent-size 4
  "*Amount to indent by in idm-mode")

(defun idm-indent-line (&optional whole-exp)
  "Indent current line"
  (interactive)
  (let ((indent (idm-indent-level))
	(pos (- (point-max) (point))) beg)
    (beginning-of-line)
    (setq beg (point))
    (skip-chars-forward " \t")
    (if (zerop (- indent (current-column)))
	nil
      (delete-region beg (point))
      (indent-to indent))
    (if (> (- (point-max) pos) (point))
	(goto-char (- (point-max) pos)))
    ))

(defun idm-is-block-start ()
  "Check whether the point is the start of a block"
  (and (looking-at "\\([{]\\|\\<\\(case\\|for\\|if\\|while\\)\\>\\)")
       (not (looking-at ".*\\([}]\\|\\<\\(endcase\\|endfor\\|endif\\|endwhile\\)\\>\\)$"))))

(defun idm-is-block-end ()
  "Check whether the point is the start of a block"
  (looking-at "\\([}]\\|\\<\\(endcase\\|endfor\\|endif\\|endwhile\\)\\>\\)"))

(defun idm-goto-block-start ()
  "Goto the start of the block containing point (or beginning of buffer if not
   in a block)"
  (let ((l 1))
    (while (and (not (bobp)) (> l 0))
      (forward-line -1)
      (unless (bobp)
        (skip-chars-forward " \t")
        (cond
         ((idm-is-block-end) (setq l (1+ l)))
         ((idm-is-block-start) (setq l (1- l)))
         ))
    )
  (bobp))
  )

(defun idm-indent-level ()
  (save-excursion
    (let ((p (point))
          ind 
          (is-block-start nil)
          (is-block-end nil)
          (is-case-item nil)
          (is-outdent-item nil))
      ;; see what's on this line
      (beginning-of-line)
      (skip-chars-forward " \t")
      (setq is-block-end (idm-is-block-end))
      (setq is-block-start (idm-is-block-start))
      (setq is-case-item (looking-at "\\<\\(in\\|otherwise\\)\\>"))
      (setq is-outdent-item (looking-at "\\<\\(else\\|elseif\\|then\\|do\\)\\>"))

      ;; goto start of current block (0 if at top level) and
      ;; set ind to indent size more than that indentation
      (if (idm-goto-block-start)
          (setq ind 0)
        (setq ind (+ (current-indentation) idm-indent-size)))
      
      ;; indentation of this line is idm-indent-size more than that of the 
      ;; previous block, unless this line is a case item or else
      (cond (is-block-start  ind)
            (is-block-end    (- ind idm-indent-size))
            (is-case-item    (- ind idm-indent-size))
            (is-outdent-item (- ind idm-indent-size))
            (t               ind)
            )
      )))

(defun idm-electric-newline ()
  "Re-indent the current line, add a new line and indent that new line."
  (interactive "*")
  (idm-indent-line)
  (newline)
  (idm-indent-line))

(defvar idm-mode-syntax-table
  (let ((idm-mode-syntax-table (make-syntax-table)))
	
    ;; _ can be symbol constituent.
    (modify-syntax-entry ?_ "_" idm-mode-syntax-table)
	
    ;; Comments are !! to end of line.
    (modify-syntax-entry ?! ". 12" idm-mode-syntax-table)
    (modify-syntax-entry ?\n ">" idm-mode-syntax-table)

    ;; Strings begin and end with ". \ is the in-string escape character.
    (modify-syntax-entry ?\" "\""   idm-mode-syntax-table)
    (modify-syntax-entry ?\\ "\\"   idm-mode-syntax-table)

    idm-mode-syntax-table)
  "Syntax table for idm-mode")
  
(defun idm-mode ()
  (interactive)
  (kill-all-local-variables)
  (use-local-map idm-mode-map)
  (set-syntax-table idm-mode-syntax-table)
  ;; Set up font-lock
  (set (make-local-variable 'font-lock-defaults) '(idm-font-lock-keywords))
  ;; Register our indentation function
  (set (make-local-variable 'indent-line-function) 'idm-indent-line)  
  (setq major-mode 'idm-mode)
  (setq mode-name "IDM")
  (run-hooks 'idm-mode-hook))

(provide 'idm-mode)

;;; idm-mode.el ends here



